! lv_filters_mod.f90
! Authors: Joe Jennings, Ettore Biondi
! Last modified: July 16, 2015
!
! A module for computing a linearly varying filter

! TODO:
! 1. Add a mask to the object (see GEE, Fomel - helix) 
!    as part of the initialization, to detect when off the grid
! 2. QC the js on a different example
! 3. 2D case

module lv_filters_mod

  ! lag(number of dimensions, number of lags)
  ! b  (number of dimensions, number of filters-1) <- should be number of patches
  ! e  (number of dimensions, number of filters-1)
  implicit none
  type lv_filter
    integer, pointer :: lag(:,:), b(:,:), e(:,:)
  end type lv_filter

  ! coef(number of lags, number of filters)
  type, extends ( lv_filter ) :: lv_filter1d
    real,    pointer :: coef1(:,:)
    contains
      procedure, private :: filter_init1d 
      procedure, public  :: filter_close1d 
  end type lv_filter1d

  interface lv_filter1d
    procedure :: filter_create1d
  end interface

  type, extends ( lv_filter ) :: lv_filter2d
    real,    pointer :: coef2(:,:,:)
    logical, pointer :: mis  (:,:)
    contains
      ! TODO: perhaps a routine for computing mis here?
      procedure, private :: filter_init2d 
      procedure, public  :: filter_close2d
  end type lv_filter2d

  interface lv_filter2d
    procedure :: filter_create2d
  end interface

  type, extends ( lv_filter ) :: lv_filter3d
    real,    pointer :: coef3(:,:,:,:)
    logical, pointer :: mis3  (:,:,:)
    contains
      ! TODO: perhaps a routine for computing mis here?
      procedure, private :: filter_init3d 
      procedure, public  :: filter_close3d
  end type lv_filter3d

  interface lv_filter3d
    procedure :: filter_create3d
  end interface


contains

  subroutine filter_init1d(this, nf, nlag, lag_in)
    optional ::                            lag_in
    integer, intent(in) :: nf, nlag
    class(lv_filter1d), intent(inout)  :: this
    integer :: lag_in(:,:)
    integer :: i, nf_max, nf_total

    allocate(this%b(1,nf-1), this%e(1,nf-1), &
             this%coef1(nlag,nf), this%lag(1,nlag))
    
    this%coef1 = 0.
    if(present(lag_in)) then
      this%lag = lag_in
    else
      do i = 0,nlag-1
        this%lag(1,i+1) = i
      enddo
    endif
      
  end subroutine filter_init1d

  !Destructor
  subroutine filter_close1d(this)
    class(lv_filter1d) :: this
    deallocate(this%b, this%e, this%coef1, this%lag)
  end subroutine filter_close1d

  ! Constructor
  ! Allocates memory for the filter coefficients,
  ! the position arrays for the filters
  ! and the lag arrays
 function filter_create1d(ndat, j, nlag, lag_in) result(this)
   optional ::                           lag_in
    integer, intent(in) :: ndat, j, nlag
    integer  :: lag_in(:,:)
    type(lv_filter1d)   :: this

    integer, allocatable :: b1(:), e1(:)
    integer :: nfilt

    ! Computing the number of filters on each axis
    nfilt= int(ceiling(real(ndat-2)/real(j)) + 1)

    allocate(b1(nfilt-1), e1(nfilt-1))

    ! Allocating memory for the coefficients and
    ! beginning and ending arrays
    if(present(lag_in)) then
      call this%filter_init1d(nfilt,nlag,lag_in)
    else
      call this%filter_init1d(nfilt,nlag)
    endif

    ! Computing the optimal sizes for the filters
    call find_optimal_sizes(ndat,j,nfilt,b1,e1)

    this%b(1,:) = b1
    this%e(1,:) = e1
      
    ! Taking care of the endpoint
    this%e(1,nfilt-1) = this%e(1,nfilt-1) - this%lag(1,nlag)
    deallocate(b1,e1)

  end function filter_create1d

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine filter_init2d(this, ndat, nf, lag_in)
    class(lv_filter2d), intent(inout) :: this
    integer, intent(in) :: nf(:), lag_in(:,:), ndat(:)

    integer :: num_patches, i, j, k, ndim, lagdim1, lagdim2

    lagdim1 = size(lag_in,1)
    lagdim2 = size(lag_in,2)

    ndim = size(nf)

    num_patches = (nf(1)-1)
    do i = 2,size(nf)
      num_patches = num_patches * (nf(i)-1)
    enddo

    !TODO: allocate 2D mis array here
    allocate(this%b(ndim,num_patches), this%e(ndim,num_patches), &
             this%lag(ndim,lagdim2), this%coef2(lagdim2,nf(1),nf(2)), this%mis(ndat(1),ndat(2)))

    this%coef2 = 0.
    this%lag = lag_in

  end subroutine filter_init2d

  subroutine filter_close2d(this)
    class(lv_filter2d), intent(inout) :: this
    deallocate(this%b, this%e, this%coef2, this%lag)
  end subroutine filter_close2d

  function filter_create2d(ndat,  j, lag_in) result(this)
    integer, intent(in) :: lag_in(:,:), ndat(:), j(:)
    type(lv_filter2d) :: this

    integer, allocatable :: nfilt(:), b1(:), e1(:), b2(:), e2(:)
    integer              :: i1, i2, i3

    allocate(nfilt(size(ndat)))
    
    ! Number of filters per axis
    do i1 = 1,size(ndat)
      nfilt(i1) = int(ceiling(real(ndat(i1)-2)/real(j(i1))) + 1)
    enddo

    allocate(b1(nfilt(1)-1), e1(nfilt(1)-1), b2(nfilt(2)-1), e2(nfilt(2)-1))

    ! Allocating memory
    call this%filter_init2d(ndat,nfilt,lag_in)

    ! Computing the filter sizes in each dimension
    call find_optimal_sizes(ndat(1), j(1), nfilt(1), b1, e1)
    call find_optimal_sizes(ndat(2), j(2), nfilt(2), b2, e2)

    ! Writing b and e to 2D arrays
    ! and taking care of the endpoints
    i3 = 1
    do i1 = 1,nfilt(2)-1
      do i2 = 1,nfilt(1)-1
        if(e1(i2) == ndat(1)) then
          this%e(1,i3) = e1(i2) - maxval(this%lag(1,:))
        else
          this%e(1,i3) = e1(i2)
        endif
        if(e2(i1) == ndat(2)) then
          this%e(2,i3) = e2(i1) - maxval(this%lag(2,:))
        else
          this%e(2,i3) = e2(i1)
        endif
        if (b1(i2) == 1) then
        this%b(1,i3) = b1(i2) - minval(this%lag(1,:))
        else 
        this%b(1,i3) = b1(i2)
        end if
        if(b2(i1) == 1) then
        this%b(2,i3) = b2(i1) - minval(this%lag(2,:))
        else
        this%b(2,i3) = b2(i1)
        endif
        i3 = i3 + 1
      enddo
    enddo

    deallocate(nfilt, b1, e1, b2, e2)

  end function filter_create2d

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine filter_init3d(this, ndat, nf, lag)
    class(lv_filter3d), intent(inout) :: this
    integer, intent(in) :: nf(:), ndat(:), lag(3) ! lag(i) = nb of lags wanted along axis i

    integer :: num_patches, i, j, k, ndim, nlag
    integer :: lag1_min, lag1_max, idx
    integer, allocatable ::lag_in(:,:)
    if (mod(lag(1),2)==0) then
      lag1_min = -lag(1)/2
      lag1_max = lag(1)/2 - 1 
    else
      lag1_min = -(lag(1) - 1)/2
      lag1_max = - lag1_min
    end if
    
    nlag = lag(1)*lag(2)*lag(3)+lag1_min
    ndim = size(nf)

    num_patches = (nf(1)-1)
    do i = 2,size(nf)
      num_patches = num_patches * (nf(i)-1)
    enddo

    allocate(this%b(ndim,num_patches), this%e(ndim,num_patches), &
             this%lag(ndim,nlag), this%coef3(nlag,nf(1),nf(2),nf(3)), &
             this%mis3(ndat(1),ndat(2),ndat(3)),lag_in(ndim,lag(1)*lag(2)*lag(3)))

	do k = 1, lag(3)
	  do j = 1, lag(2)
	    do i = 1, lag(1)
	    idx = i + (j-1)*lag(1)+ (k-1)*lag(1)*lag(2)
	      lag_in(1, idx) = lag1_min + i - 1
	      lag_in(2, idx) = j - 1
	      lag_in(3, idx) = k - 1
	    end do
	  end do
	end do    

    this%coef3 = 0.
    this%lag = lag_in(:, 1-lag1_min:lag(1)*lag(2)*lag(3))
  end subroutine filter_init3d

  subroutine filter_close3d(this)
    class(lv_filter3d), intent(inout) :: this
    deallocate(this%b, this%e, this%coef3, this%lag,this%mis3)
  end subroutine filter_close3d
  
  function filter_create3d(ndat,  j, lag) result(this)
    integer, intent(in) :: lag(3), ndat(:), j(:)
    type(lv_filter3d)   :: this

    integer, allocatable :: nfilt(:), b1(:), e1(:), b2(:), e2(:),b3(:),e3(:)
    integer              :: i1, i2, i3, k

    allocate(nfilt(size(ndat)))
    
    ! Number of filters per axis
    do i1 = 1,size(ndat)
      nfilt(i1) = int(ceiling(real(ndat(i1)-2)/real(j(i1))) + 1)
    enddo

    allocate(b1(nfilt(1)-1), e1(nfilt(1)-1), b2(nfilt(2)-1), e2(nfilt(2)-1), &
             b3(nfilt(3)-1), e3(nfilt(3)-1))

    ! Allocating memory
    call this%filter_init3d(ndat,nfilt,lag)

    ! Computing the filter sizes in each dimension
    call find_optimal_sizes(ndat(1), j(1), nfilt(1), b1, e1)
    call find_optimal_sizes(ndat(2), j(2), nfilt(2), b2, e2)
    call find_optimal_sizes(ndat(3), j(3), nfilt(3), b3, e3)

    ! Writing b and e to 2D arrays
    ! and taking care of the endpoints
    k = 1
    do i3 = 1,nfilt(3)-1
     do i2 = 1,nfilt(2)-1
      do i1 = 1,nfilt(1)-1
        if(e1(i1) == ndat(1)) then
          this%e(1,k) = e1(i1) - maxval(this%lag(1,:))
        else
          this%e(1,k) = e1(i1)
        endif
        if(e2(i2) == ndat(2)) then
          this%e(2,k) = e2(i2) - maxval(this%lag(2,:))
        else
          this%e(2,k) = e2(i2)
        endif
        if(e3(i3) == ndat(3)) then
          this%e(3,k) = e3(i3) - maxval(this%lag(3,:))
        else
          this%e(3,k) = e3(i3)
        endif
        if (b1(i1) == 1) then
        this%b(1,k) = b1(i1) - minval(this%lag(1,:))
        else 
        this%b(1,k) = b1(i1)
        end if
        if(b2(i2) == 1) then
        this%b(2,k) = b2(i2) - minval(this%lag(2,:))
        else
        this%b(2,k) = b2(i2)
        endif
        if (b3(i3) == 1) then
        this%b(3,k) = b3(i3) - minval(this%lag(3,:))
        else 
        this%b(3,k) = b3(i3)
        end if
        k = k + 1
      enddo
	 enddo
    enddo

    deallocate(nfilt, b1, e1, b2, e2,b3,e3)

  end function filter_create3d

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine find_optimal_sizes(nd,j,nf,b,e)
    integer :: nd,j,nf,b(:),e(:)
    integer :: last,space,sample,k
    space = nd
    b(1)= 0
    do k=0,nf-2
      sample = ceiling((real(space)/real(nf-k-1)))
      e(k+1) = sample+b(k+1)-1 
      if(k /= nf-2) b(k+2)=e(k+1)+1
      space = space - sample
    end do
    b=b+1
    e=e+1
  end subroutine find_optimal_sizes

end module lv_filters_mod
